from abc import ABC

from fhirpy.base.resource import BaseResource, BaseReference
from .base import (
    SyncClient,
    AsyncClient,
    SyncSearchSet,
    AsyncSearchSet,
    SyncResource,
    AsyncResource,
    SyncReference,
    AsyncReference,
)


class SyncFHIRSearchSet(SyncSearchSet):
    pass


class AsyncFHIRSearchSet(AsyncSearchSet):
    pass


class BaseFHIRResource(BaseResource, ABC):
    def is_reference(self, value):
        if not isinstance(value, dict):
            return False

        return "reference" in value and not (
            set(value.keys())
            - {"reference", "display", "type", "identifier", "extension"}
        )


class SyncFHIRResource(BaseFHIRResource, SyncResource):
    pass


class AsyncFHIRResource(BaseFHIRResource, AsyncResource):
    pass


class BaseFHIRReference(BaseReference, ABC):
    @property
    def reference(self):
        return self["reference"]

    @property
    def id(self):
        """
        Returns id if reference specifies to the local resource
        """
        if self.is_local:
            return self.reference.split("/", 1)[1]

    @property
    def resource_type(self):
        """
        Returns resource type if reference specifies to the local resource
        """
        if self.is_local:
            return self.reference.split("/", 1)[0]

    @property
    def is_local(self):
        return self.reference.count("/") == 1


class SyncFHIRReference(BaseFHIRReference, SyncReference):
    pass


class AsyncFHIRReference(BaseFHIRReference, AsyncReference):
    pass


class SyncFHIRClient(SyncClient):
    searchset_class = SyncFHIRSearchSet
    resource_class = SyncFHIRResource

    def reference(self, resource_type=None, id=None, reference=None, **kwargs):
        if resource_type and id:
            reference = "{0}/{1}".format(resource_type, id)

        if not reference:
            raise TypeError(
                "Arguments `resource_type` and `id` or `reference` " "are required"
            )
        return SyncFHIRReference(self, reference=reference, **kwargs)


class AsyncFHIRClient(AsyncClient):
    searchset_class = AsyncFHIRSearchSet
    resource_class = AsyncFHIRResource

    def reference(self, resource_type=None, id=None, reference=None, **kwargs):
        if resource_type and id:
            reference = "{0}/{1}".format(resource_type, id)

        if not reference:
            raise TypeError(
                "Arguments `resource_type` and `id` or `reference` " "are required"
            )
        return AsyncFHIRReference(self, reference=reference, **kwargs)



import copy
import datetime
from abc import ABC, abstractmethod
from collections import defaultdict

import pytz

from fhirpy.base.resource import BaseResource, BaseReference
from fhirpy.base.utils import chunks, encode_params
from fhirpy.base.exceptions import InvalidResponse

FHIR_DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
FHIR_DATE_FORMAT = "%Y-%m-%d"


def format_date_time(date: datetime.datetime):
    return pytz.utc.normalize(date).strftime(FHIR_DATE_TIME_FORMAT)


def format_date(date: datetime.date):
    return date.strftime(FHIR_DATE_FORMAT)


def transform_param(param: str):
    """
    >>> transform_param('general_practitioner')
    'general-practitioner'
    """
    if param[0] == "_" or param[0] == ".":
        # Don't correct _id, _has, _include, .effectiveDate and etc.
        return param

    return param.replace("_", "-")


def transform_value(value):
    """
    >>> transform_value(datetime.datetime(2019, 1, 1, tzinfo=pytz.utc))
    '2019-01-01T00:00:00Z'

    >>> transform_value(datetime.date(2019, 1, 1))
    '2019-01-01'

    >>> transform_value(True)
    'true'
    """
    if isinstance(value, datetime.datetime):
        return format_date_time(value)
    if isinstance(value, datetime.date):
        return format_date(value)
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, (BaseReference, BaseResource)):
        return value.reference
    return value


class Raw:
    kwargs = {}

    def __init__(self, **kwargs):
        self.kwargs = kwargs


def SQ(*args, **kwargs):
    """
    Builds search query

    >>> dict(SQ(general_practitioner='prid'))
    {'general-practitioner': ['prid']}

    >>> dict(SQ(patient__Patient__name='John'))
    {'patient:Patient.name': ['John']}

    >>> dict(SQ(patient__Patient__birth_date__ge='2000'))
    {'patient:Patient.birth-date': ['ge2000']}

    >>> dict(SQ(patient__Patient__general_practitioner__Organization__name='Name'))
    {'patient:Patient.general-practitioner:Organization.name': ['Name']}

    >>> dict(SQ(patient__Patient__general_practitioner__name='Name'))
    {'patient:Patient.general-practitioner.name': ['Name']}

    >>> dict(SQ(based_on__instantiates_canonical='PlanDefinition/id'))
    {'based-on.instantiates-canonical': ['PlanDefinition/id']}

    >>> dict(SQ(period__ge='2018', period__lt='2019'))
    {'period': ['ge2018', 'lt2019']}

    >>> dict(SQ(text__contains='test'))
    {'text:contains': ['test']}

    >>> dict(SQ(url__not_in='http://loinc.org'))
    {'url:not-in': ['http://loinc.org']}

    >>> dict(SQ(name='family1,family2'))
    {'name': ['family1,family2']}

    >>> dict(SQ(status__not=['failed', 'completed']))
    {'status:not': ['failed', 'completed']}

    >>> dict(SQ(active=True))
    {'active': ['true']}

    >>> dict(SQ(**{'_has:Person:link:id': 'id'}))
    {'_has:Person:link:id': ['id']}

    >>> dict(SQ(**{'.effectiveDate.start$gt': '2019'}))
    {'.effectiveDate.start$gt': ['2019']}

    >>> dict(SQ(_lastUpdated__gt=2019))
    {'_lastUpdated': ['gt2019']}

    >>> dict(SQ(Raw(**{'_has:Person:link:id': 'id'})))
    {'_has:Person:link:id': ['id']}

    """
    param_ops = [
        "contains",
        "exact",
        "missing",
        "not",
        "below",
        "above",
        "in",
        "not_in",
        "text",
        "of_type",
    ]
    value_ops = ["eq", "ne", "gt", "ge", "lt", "le", "sa", "eb", "ap"]

    res = defaultdict(list)
    for key, value in kwargs.items():
        value = value if isinstance(value, list) else [value]
        value = [transform_value(sub_value) for sub_value in value]

        key_parts = key.split("__")

        op = None
        if key_parts[-1] in value_ops or key_parts[-1] in param_ops:
            # The operator is always the last part,
            # e.g., birth_date__ge or patient__Patient__birth_date__ge
            op = key_parts[-1]
            key_parts = key_parts[:-1]

        param = key_parts[0]
        for part in key_parts[1:]:
            # Resource type always starts with upper first letter
            is_resource_type = part[0].isupper()

            param += ":" if is_resource_type else "."
            param += part

        if op:
            if op in param_ops:
                param = "{0}:{1}".format(param, transform_param(op))
            elif op in value_ops:
                value = ["{0}{1}".format(op, sub_value) for sub_value in value]
        res[transform_param(param)].extend(value)

    for arg in args:
        if isinstance(arg, Raw):
            for key, value in arg.kwargs.items():
                value = value if isinstance(value, list) else [value]
                res[key].extend(value)
        else:
            raise ValueError("Can't handle args without Raw() wrapper")

    return res


class AbstractSearchSet(ABC):
    client = None
    resource_type = None
    params = None

    def __init__(self, client, resource_type, params=None):
        self.client = client
        self.resource_type = resource_type
        self.params = defaultdict(list, params or {})

    def _dict_to_resource(self, data):
        return self.client.resource(data['resourceType'], **data)

    @abstractmethod  # pragma: no cover
    def fetch(self):
        pass

    @abstractmethod  # pragma: no cover
    def fetch_raw(self):
        pass

    @abstractmethod  # pragma: no cover
    def fetch_all(self):
        pass

    @abstractmethod  # pragma: no cover
    def get(self, id):
        pass

    @abstractmethod  # pragma: no cover
    def count(self):
        pass

    @abstractmethod  # pragma: no cover
    def first(self):
        pass

    @abstractmethod
    async def get_or_create(self, resource):
        pass

    @abstractmethod
    def update(self, resource):
        pass

    @abstractmethod
    def patch(self, resource):
        pass

    def clone(self, override=False, **kwargs):
        new_params = copy.deepcopy(self.params)
        for key, value in kwargs.items():
            if not isinstance(value, list):
                value = [value]

            if override:
                new_params[key] = value
            else:
                new_params[key].extend(value)

        return self.__class__(self.client, self.resource_type, new_params)

    def elements(self, *attrs, exclude=False):
        attrs = set(attrs)
        if not exclude:
            attrs |= {"id", "resourceType"}
        attrs = [attr for attr in attrs]

        return self.clone(
            _elements="{0}{1}".format("-" if exclude else "", ",".join(attrs)),
            override=True,
        )

    def has(self, *args, **kwargs):
        if len(args) % 2 != 0:
            raise TypeError(
                "You should pass even size of arguments, for example: "
                "`.has('Observation', 'patient', "
                "'AuditEvent', 'entity', user='id')`"
            )

        key_part = ":".join(
            ["_has:{0}".format(":".join(pair)) for pair in chunks(args, 2)]
        )

        return self.clone(
            **{":".join([key_part, key]): value for key, value in SQ(**kwargs).items()}
        )

    def include(
        self,
        resource_type,
        attr=None,
        target_resource_type=None,
        *,
        recursive=False,
        iterate=False,
        reverse=False
    ):
        key_params = ["_revinclude" if reverse else "_include"]

        if iterate:
            # Added in FHIR v3.5
            key_params.append("iterate")
        if recursive:
            # Works for FHIR v3.0-3.3
            key_params.append("recursive")
        key = ":".join(key_params)

        if resource_type == "*":
            value = "*"
        else:
            if not attr:
                raise TypeError(
                    "You should provide attr " "(search parameter) argument"
                )
            value_params = [resource_type, attr]
            if target_resource_type:
                value_params.append(target_resource_type)
            value = ":".join(value_params)

        return self.clone(**{key: value})

    def revinclude(
        self,
        resource_type,
        attr=None,
        target_resource_type=None,
        *,
        recursive=False,
        iterate=False
    ):
        return self.include(
            resource_type,
            attr=attr,
            target_resource_type=target_resource_type,
            recursive=recursive,
            iterate=iterate,
            reverse=True,
        )

    def search(self, *args, **kwargs):
        return self.clone(**SQ(*args, **kwargs))

    def limit(self, limit):
        return self.clone(_count=limit, override=True)

    def sort(self, *keys):
        sort_keys = ",".join(keys)
        return self.clone(_sort=sort_keys, override=True)

    def __str__(self):  # pragma: no cover
        return "<{0} {1}?{2}>".format(
            self.__class__.__name__, self.resource_type, encode_params(self.params)
        )

    def __repr__(self):  # pragma: no cover
        return self.__str__()

    def _get_bundle_resources(self, bundle_data):
        bundle_resource_type = bundle_data.get("resourceType", None)

        if bundle_resource_type != "Bundle":
            raise InvalidResponse(
                "Expected to receive Bundle "
                "but {0} received".format(bundle_resource_type)
            )

        resources_data = [res["resource"] for res in bundle_data.get("entry", [])]

        resources = []
        for data in resources_data:
            resource = self._dict_to_resource(data)
            if resource.resource_type == self.resource_type:
                resources.append(resource)
        return resources



from .lib import SyncFHIRClient, AsyncFHIRClient

__title__ = "fhir-py"
__version__ = "1.4.1"
__author__ = "beda.software"
__license__ = "None"
__copyright__ = "Copyright 2023 beda.software"

# Version synonym
VERSION = __version__



import json
from enum import Enum


class BaseFHIRError(Exception):
    pass


class ResourceNotFound(BaseFHIRError):
    pass


class InvalidResponse(BaseFHIRError):
    pass


class AuthorizationError(BaseFHIRError):
    pass


class OperationOutcome(BaseFHIRError):
    pass


class MultipleResourcesFound(BaseFHIRError):
    pass


class IssueType(Enum):
    """Some Issue types from https://www.hl7.org/fhir/valueset-issue-type.html"""

    invalid = "invalid"  # Content invalid against the specification or a profile.
    required = "required"  # A required element is missing.
    forbidden = "forbidden"  # The user does not have the rights to perform this action.
    not_found = "not-found"  # The reference provided was not found.
    exception = "exception"  # An unexpected internal error has occurred.
    informational = "informational"  # A message unrelated to the processing success of the completed operation (examples of the latter include things like reminders of password expiry, system maintenance times, etc.).


class IssueSeverity(Enum):
    fatal = "fatal"
    error = "error"
    warning = "warning"
    information = "information"


class OperationOutcome(BaseFHIRError):
    def __init__(
        self,
        reason=None,
        *,
        resource=None,
        severity=IssueSeverity.fatal.value,
        code=IssueType.invalid.value
    ):
        self.resource = resource or {
            "resourceType": "OperationOutcome",
            "issue": [
                {
                    "severity": severity,
                    "code": code,
                    "diagnostics": reason or "Error",
                }
            ],
            "text": {
                "status": "generated",
                "div": reason or "Something went wrong",
            },
        }
        super().__init__(json.dumps(self.resource, indent=2))



import reprlib
from urllib.parse import urlencode, quote, parse_qs, urlparse
from yarl import URL


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def get_by_path(self, path, default=None):
        keys = parse_path(path)
        return get_by_path(self, keys, default)


class SearchList(list):
    def get_by_path(self, path, default=None):
        keys = parse_path(path)
        return get_by_path(self, keys, default)


def chunks(l, n):
    """
    Yield successive n-sized chunks from l

    >>> list(chunks([1, 2, 3, 4], 2))
    [[1, 2], [3, 4]]
    """
    for i in range(0, len(l), n):
        yield l[i : i + n]


def unique_everseen(seq):
    """
    >>> unique_everseen(['1', '2', '3', '1', '2'])
    ['1', '2', '3']
    """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def encode_params(params):
    """
    >>> encode_params({'status:not': ['active', 'entered-in-error']})
    'status:not=active&status:not=entered-in-error'

    >>> encode_params({'status': ['active,waitlist']})
    'status=active,waitlist'

    >>> encode_params({'status': 'active,waitlist'})
    'status=active,waitlist'

    >>> encode_params({'_format': ['json', 'json']})
    '_format=json'

    >>> encode_params(None)
    ''
    """
    params = params or {}
    return urlencode(
        {
            k: unique_everseen(v) if isinstance(v, list) else [v]
            for k, v in params.items()
        },
        doseq=True,
        safe=":,",
        quote_via=quote,
    )


def parse_pagination_url(url):
    """
    Parses Bundle.link pagination url and returns path and params

    >>> parse_pagination_url('/Patient?_count=100&name=ivan&name=petrov')
    ('/Patient', {'_count': ['100'], 'name': ['ivan', 'petrov']})
    """
    if URL(url).is_absolute():
        return url, None
    parsed = urlparse(url)
    params = parse_qs(parsed.query)
    path = parsed.path

    return path, params


def convert_values(data, fn):
    """
    Recursively converts data values with `fn`
    which must return tuple of (converted data, stop flag).
    Conversion will be stopped for this branch if stop flag is True

    >>> convert_values({}, lambda x: (x, False))
    {}

    >>> convert_values([], lambda x: (x, False))
    []

    >>> convert_values('str', lambda x: (x, False))
    'str'

    >>> convert_values(
    ... [{'key1': [1, 2]}, {'key2': [3, 4]}],
    ... lambda x: (x + 1, False) if isinstance(x, int) else (x, False))
    [{'key1': [2, 3]}, {'key2': [4, 5]}]

    >>> convert_values(
    ... [{'replaceable': True}, {'replaceable': False}],
    ... lambda x: ('replaced', False)
    ...     if isinstance(x, dict) and x.get('replaceable', False)
    ...     else (x, False))
    ['replaced', {'replaceable': False}]
    """

    data, stop = fn(data)

    if stop:
        return data

    if isinstance(data, list):
        return SearchList(convert_values(x, fn) for x in data)
    if isinstance(data, dict):
        return AttrDict({key: convert_values(value, fn) for key, value in dict.items(data)})
    return data


def parse_path(path):
    """
    >>> parse_path(['path', 'to', 0, 'element'])
    ['path', 'to', 0, 'element']

    >>> parse_path('path.to.0.element')
    ['path', 'to', 0, 'element']
    """
    if isinstance(path, str):
        return [int(key) if key.isdigit() else key for key in path.split(".")]
    elif isinstance(path, list):
        return path
    else:  # pragma: no cover
        raise TypeError("Path must be or a dotted string or a list")


def get_by_path(data, path, default=None):
    """
    >>> get_by_path({'key': 'value'}, ['key'])
    'value'

    >>> get_by_path({'key': [{'nkey': 'nvalue'}]}, ['key', 0, 'nkey'])
    'nvalue'

    >>> get_by_path({
    ...     'key': [
    ...         {'test': 'test0', 'nkey': 'zero'},
    ...         {'test': 'test1', 'nkey': 'one'}
    ...     ]
    ... }, ['key', {'test': 'test1'}, 'nkey'])
    'one'

    >>> get_by_path({'a': 1}, ['b'], 0)
    0

    >>> get_by_path({'a': {'b': None}}, ['a', 'b'], 0) is None
    True

    >>> get_by_path({'a': {'b': None}}, ['a', 'b', 'c'], 0)
    0
    """
    assert isinstance(path, list), "Path must be a list"

    rv = data
    try:
        for key in path:
            if rv is None:
                return default

            if isinstance(rv, list):
                if isinstance(key, int):
                    rv = rv[key]
                elif isinstance(key, dict):
                    matched_index = -1
                    for index, item in enumerate(rv):
                        if all([item.get(k, None) == v for k, v in key.items()]):
                            matched_index = index
                            break
                    if matched_index == -1:
                        rv = None
                    else:
                        rv = rv[matched_index]
                else:  # pragma: no cover
                    raise TypeError(
                        "Can not lookup by {0} in list. "
                        "Possible lookups are by int or by dict.".format(
                            reprlib.repr(key)
                        )
                    )
            else:
                rv = rv[key]

        return rv
    except (IndexError, KeyError, AttributeError):
        return default


def set_by_path(obj, path, value):
    cursor = obj
    last_part = path.pop()

    for index, part in enumerate(path):
        if isinstance(cursor, dict) and part not in cursor:
            nextpart = (path + [last_part])[index + 1]
            try:
                nnextpart = (path + [last_part])[index + 2]
            except IndexError:
                nnextpart = ""

            if isinstance(nextpart, int):
                cursor[part] = [[] if isinstance(nnextpart, int) else {}]
            else:
                cursor[part] = {}

        cursor = cursor[part]
    cursor[last_part] = value


def remove_prefix(s, prefix):
    return s[len(prefix) :] if s.startswith(prefix) else s



import json
import copy
import logging
import warnings
from abc import ABC, abstractmethod
from json import JSONDecodeError

import aiohttp
import requests

from yarl import URL
from fhirpy.base.searchset import AbstractSearchSet
from fhirpy.base.resource import BaseResource, BaseReference
from fhirpy.base.utils import (
    AttrDict,
    encode_params,
    get_by_path,
    parse_pagination_url,
    remove_prefix,
)
from fhirpy.base.exceptions import (
    ResourceNotFound,
    OperationOutcome,
    InvalidResponse,
    MultipleResourcesFound,
)


class AbstractClient(ABC):
    url = None
    authorization = None
    extra_headers = None

    def __init__(self, url, authorization=None, extra_headers=None):
        self.url = url
        self.authorization = authorization
        self.extra_headers = extra_headers

    def __str__(self):  # pragma: no cover
        return "<{0} {1}>".format(self.__class__.__name__, self.url)

    def __repr__(self):  # pragma: no cover
        return self.__str__()

    @property  # pragma: no cover
    @abstractmethod
    def searchset_class(self):
        pass

    @property  # pragma: no cover
    @abstractmethod
    def resource_class(self):
        pass

    @abstractmethod  # pragma: no cover
    def reference(self, resource_type=None, id=None, reference=None, **kwargs):
        pass

    def resource(self, resource_type=None, **kwargs):
        if resource_type is None:
            raise TypeError("Argument `resource_type` is required")

        return self.resource_class(self, resource_type=resource_type, **kwargs)

    def resources(self, resource_type):
        return self.searchset_class(self, resource_type=resource_type)

    @abstractmethod  # pragma: no cover
    def execute(self, path, method=None, **kwargs):
        pass

    @abstractmethod  # pragma: no cover
    def _do_request(self, method, path, data=None, params=None, returning_status=False):
        pass

    @abstractmethod  # pragma: no cover
    def _fetch_resource(self, path, params=None):
        pass

    def _build_request_headers(self):
        headers = {"Accept": "application/fhir+json"}

        if self.authorization:
            headers["Authorization"] = self.authorization

        if self.extra_headers is not None:
            headers = {**headers, **self.extra_headers}

        return headers

    def _build_request_url(self, path, params):
        if URL(path).is_absolute():
            if self.url in path:
                return path
            else:
                raise ValueError(
                    f'Request url "{path}" does not contain base url "{self.url}"'
                    " (possible security issue)"
                )
        path = path.lstrip("/")
        base_url_path = URL(self.url).path.lstrip("/") + "/"
        path = remove_prefix(path, base_url_path)
        params = params or {}

        return f'{self.url.rstrip("/")}/{path.lstrip("/")}?{encode_params(params)}'


class AsyncClient(AbstractClient, ABC):
    aiohttp_config = None

    def __init__(
        self, url, authorization=None, extra_headers=None, aiohttp_config=None
    ):
        self.aiohttp_config = aiohttp_config or {}

        super().__init__(url, authorization, extra_headers)

    async def execute(self, path, method="post", **kwargs):
        return await self._do_request(method, path, **kwargs)

    async def _do_request(self, method, path, data=None, params=None, returning_status=False):
        headers = self._build_request_headers()
        url = self._build_request_url(path, params)
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.request(
                method, url, json=data, **self.aiohttp_config
            ) as r:
                if 200 <= r.status < 300:
                    data = await r.text()
                    r_data = json.loads(data, object_hook=AttrDict) if data else None
                    return (r_data, r.status) if returning_status else r_data

                if r.status == 404 or r.status == 410:
                    raise ResourceNotFound(await r.text())

                if r.status == 412:
                    raise MultipleResourcesFound(await r.text())

                data = await r.text()
                try:
                    parsed_data = json.loads(data)
                    if parsed_data["resourceType"] == "OperationOutcome":
                        raise OperationOutcome(resource=parsed_data)
                    raise OperationOutcome(reason=data)
                except (KeyError, JSONDecodeError):
                    raise OperationOutcome(reason=data)

    async def _fetch_resource(self, path, params=None):
        return await self._do_request("get", path, params=params)


class SyncClient(AbstractClient, ABC):
    requests_config = None

    def __init__(
        self, url, authorization=None, extra_headers=None, requests_config=None
    ):
        self.requests_config = requests_config or {}

        super().__init__(url, authorization, extra_headers)

    def execute(self, path, method="post", **kwargs):
        return self._do_request(method, path, **kwargs)

    def _do_request(self, method, path, data=None, params=None, returning_status=False):
        headers = self._build_request_headers()
        url = self._build_request_url(path, params)
        r = requests.request(
            method, url, json=data, headers=headers, **self.requests_config
        )

        if 200 <= r.status_code < 300:
            r_data = json.loads(r.content.decode(), object_hook=AttrDict) if r.content else None
            return (r_data, r.status_code) if returning_status else r_data

        if r.status_code == 404 or r.status_code == 410:
            raise ResourceNotFound(r.content.decode())

        if r.status_code == 412:
            raise MultipleResourcesFound(r.content.decode())

        data = r.content.decode()
        try:
            parsed_data = json.loads(data)
            if parsed_data["resourceType"] == "OperationOutcome":
                raise OperationOutcome(resource=parsed_data)
            raise OperationOutcome(reason=data)
        except (KeyError, JSONDecodeError):
            raise OperationOutcome(reason=data)

    def _fetch_resource(self, path, params=None):
        return self._do_request("get", path, params=params)


class SyncSearchSet(AbstractSearchSet, ABC):
    def fetch(self):
        bundle_data = self.client._fetch_resource(self.resource_type, self.params)
        resources = self._get_bundle_resources(bundle_data)
        return resources

    def fetch_raw(self):
        data = self.client._fetch_resource(self.resource_type, self.params)
        data_resource_type = data.get("resourceType", None)

        if data_resource_type == "Bundle":
            for item in data["entry"]:
                item.resource = self._dict_to_resource(item.resource)

        return data

    def fetch_all(self):
        return list([x for x in self])

    def get(self, id=None):
        searchset = self.limit(2)
        if id:
            warnings.warn(
                "parameter 'id' of method get() is deprecated "
                "and will be removed in future versions. "
                "Please use 'search(id='...').get()'",
                DeprecationWarning,
                stacklevel=2,
            )
            searchset = searchset.search(_id=id)
        res_data = searchset.fetch()
        if len(res_data) == 0:
            raise ResourceNotFound("No resources found")
        elif len(res_data) > 1:
            raise MultipleResourcesFound("More than one resource found")
        resource = res_data[0]
        return self._dict_to_resource(resource)

    def count(self):
        new_params = copy.deepcopy(self.params)
        new_params["_count"] = 0
        new_params["_totalMethod"] = "count"

        return self.client._fetch_resource(self.resource_type, params=new_params)[
            "total"
        ]

    def first(self):
        result = self.limit(1).fetch()

        return result[0] if result else None

    def get_or_create(self, resource):
        assert resource.resource_type == self.resource_type
        data, status_code = self.client._do_request("POST", self.resource_type, resource.serialize(), self.params, True)
        return data, (True if status_code == 201 else False)

    def update(self, resource):
        # TODO: Support cases where resource with id is provided
        # accordingly to the https://build.fhir.org/http.html#cond-update
        assert resource.resource_type == self.resource_type
        data, status_code = self.client._do_request("PUT", self.resource_type, resource.serialize(), self.params, True)
        return data, (True if status_code == 201 else False)

    def patch(self, resource):
        # TODO: Handle cases where resource with id is provided
        assert resource.resource_type == self.resource_type
        # TODO: Should we omit resourceType after serialization? (not to pollute history)
        return self.client._do_request("PATCH", self.resource_type, resource.serialize(), self.params)

    def __iter__(self):
        next_link = None
        while True:
            if next_link:
                bundle_data = self.client._fetch_resource(
                    *parse_pagination_url(next_link)
                )
            else:
                bundle_data = self.client._fetch_resource(
                    self.resource_type, self.params
                )
            new_resources = self._get_bundle_resources(bundle_data)
            next_link = get_by_path(bundle_data, ["link", {"relation": "next"}, "url"])

            for item in new_resources:
                yield item

            if not next_link:
                break

class AsyncSearchSet(AbstractSearchSet, ABC):
    async def fetch(self):
        bundle_data = await self.client._fetch_resource(self.resource_type, self.params)
        resources = self._get_bundle_resources(bundle_data)
        return resources

    async def fetch_raw(self):
        data = await self.client._fetch_resource(self.resource_type, self.params)
        data_resource_type = data.get("resourceType", None)

        if data_resource_type == "Bundle":
            for item in data["entry"]:
                item.resource = self._dict_to_resource(item.resource)

        return data

    async def fetch_all(self):
        return list([x async for x in self])

    async def get(self, id=None):
        searchset = self.limit(2)
        if id:
            warnings.warn(
                "parameter 'id' of method get() is deprecated "
                "and will be removed in future versions. "
                "Please use 'search(id='...').get()'",
                DeprecationWarning,
                stacklevel=2,
            )
            searchset = searchset.search(_id=id)
        res_data = await searchset.fetch()
        if len(res_data) == 0:
            raise ResourceNotFound("No resources found")
        elif len(res_data) > 1:
            raise MultipleResourcesFound("More than one resource found")
        resource = res_data[0]
        return self._dict_to_resource(resource)

    async def count(self):
        new_params = copy.deepcopy(self.params)
        new_params["_count"] = 0
        new_params["_totalMethod"] = "count"

        return (
            await self.client._fetch_resource(self.resource_type, params=new_params)
        )["total"]

    async def first(self):
        result = await self.limit(1).fetch()

        return result[0] if result else None

    async def get_or_create(self, resource):
        assert resource.resource_type == self.resource_type
        data, status_code = await self.client._do_request("POST", self.resource_type, resource.serialize(), self.params, True)
        return data, (True if status_code == 201 else False)

    async def update(self, resource):
        # TODO: Support cases where resource with id is provided
        # accordingly to the https://build.fhir.org/http.html#cond-update
        assert resource.resource_type == self.resource_type
        data, status_code = await self.client._do_request("PUT", self.resource_type, resource.serialize(), self.params, True)
        return data, (True if status_code == 201 else False)

    async def patch(self, resource):
        # TODO: Handle cases where resource with id is provided
        assert resource.resource_type == self.resource_type
        # TODO: Should we omit resourceType after serialization? (not to pollute history)
        return await self.client._do_request("PATCH", self.resource_type, resource.serialize(), self.params)

    async def __aiter__(self):
        next_link = None
        while True:
            if next_link:
                bundle_data = await self.client._fetch_resource(
                    *parse_pagination_url(next_link)
                )
            else:
                bundle_data = await self.client._fetch_resource(
                    self.resource_type, self.params
                )
            new_resources = self._get_bundle_resources(bundle_data)
            next_link = get_by_path(bundle_data, ["link", {"relation": "next"}, "url"])

            for item in new_resources:
                yield item

            if not next_link:
                break


class SyncResource(BaseResource, ABC):
    def save(self, fields=None, search_params=None):
        data = self.serialize()
        if fields:
            if not self.id:
                raise TypeError("Resource `id` is required for update operation")
            data = {key: data[key] for key in fields}
            method = "patch"
        else:
            method = "put" if self.id else "post"
        response_data = self.client._do_request(method, self._get_path(), data=data, params=search_params)
        if response_data:
            super(BaseResource, self).clear()
            super(BaseResource, self).update(
                **self.client.resource(self.resource_type, **response_data)
            )

    def create(self, **kwargs):
        self.save(search_params=kwargs)
        return self

    def update(self):
        if not self.id:
            raise TypeError("Resource `id` is required for update operation")
        self.save()
    def patch(self, **kwargs):
        super(BaseResource, self).update(**kwargs)
        self.save(fields=kwargs.keys())

    def delete(self):
        return self.client._do_request("delete", self._get_path())

    def refresh(self):
        data = self.client._do_request("get", self._get_path())
        super(BaseResource, self).clear()
        super(BaseResource, self).update(**data)

    def is_valid(self, raise_exception=False):
        data = self.client._do_request(
            "post", "{0}/$validate".format(self.resource_type), data=self.serialize()
        )
        if any(issue["severity"] in ["fatal", "error"] for issue in data["issue"]):
            if raise_exception:
                raise OperationOutcome(resource=data)
            return False
        return True

    def execute(self, operation, method="post", data=None, params=None):
        return self.client._do_request(
            method,
            "{0}/{1}".format(self._get_path(), operation),
            data=data,
            params=params,
        )


class AsyncResource(BaseResource, ABC):
    async def save(self, fields=None, search_params=None):
        data = self.serialize()
        if fields:
            if not self.id:
                raise TypeError("Resource `id` is required for update operation")
            data = {key: data[key] for key in fields}
            method = "patch"
        else:
            method = "put" if self.id else "post"

        response_data = await self.client._do_request(
            method, self._get_path(), data=data, params=search_params
        )
        if response_data:
            super(BaseResource, self).clear()
            super(BaseResource, self).update(
                **self.client.resource(self.resource_type, **response_data)
            )

    async def create(self, **kwargs):
        await self.save(search_params=kwargs)
        return self

    async def update(self):
        if not self.id:
            raise TypeError("Resource `id` is required for update operation")
        await self.save()
    async def patch(self, **kwargs):
        super(BaseResource, self).update(**kwargs)
        await self.save(fields=kwargs.keys())

    async def delete(self):
        return await self.client._do_request("delete", self._get_path())

    async def refresh(self):
        data = await self.client._do_request("get", self._get_path())
        super(BaseResource, self).clear()
        super(BaseResource, self).update(**data)

    async def to_resource(self):
        return super(AsyncResource, self).to_resource()

    async def is_valid(self, raise_exception=False):
        data = await self.client._do_request(
            "post", "{0}/$validate".format(self.resource_type), data=self.serialize()
        )
        if any(issue["severity"] in ["fatal", "error"] for issue in data["issue"]):
            if raise_exception:
                raise OperationOutcome(resource=data)
            return False
        return True

    async def execute(self, operation, method="post", **kwargs):
        return await self.client._do_request(
            method, "{0}/{1}".format(self._get_path(), operation), **kwargs
        )


class SyncReference(BaseReference, ABC):
    def to_resource(self):
        """
        Returns Resource instance for this reference
        from fhir server otherwise.
        """
        if not self.is_local:
            raise ResourceNotFound("Can not resolve not local resource")
        resource_data = self.client._do_request("get", "{0}/{1}".format(self.resource_type, self.id))
        return self._dict_to_resource(resource_data)

    def execute(self, operation, method="post", **kwargs):
        if not self.is_local:
            raise ResourceNotFound("Can not execute on not local resource")
        return self.client._do_request(
            method,
            "{0}/{1}/{2}".format(self.resource_type, self.id, operation),
            **kwargs,
        )


class AsyncReference(BaseReference, ABC):
    async def to_resource(self):
        """
        Returns Resource instance for this reference
        from fhir server otherwise.
        """
        if not self.is_local:
            raise ResourceNotFound("Can not resolve not local resource")
        resource_data = await self.client._do_request("get", "{0}/{1}".format(self.resource_type, self.id))
        return self._dict_to_resource(resource_data)

    async def execute(self, operation, method="post", **kwargs):
        if not self.is_local:
            raise ResourceNotFound("Can not execute on not local resource")
        return await self.client._do_request(
            method,
            "{0}/{1}/{2}".format(self.resource_type, self.id, operation),
            **kwargs,
        )



from abc import ABC, abstractmethod

from fhirpy.base.exceptions import ResourceNotFound
from fhirpy.base.utils import parse_path, get_by_path, convert_values


class AbstractResource(dict):
    client = None

    def __init__(self, client, **kwargs):
        self.client = client

        super(AbstractResource, self).__init__(**kwargs)

    def __eq__(self, other):
        return isinstance(other, AbstractResource) and self.reference == other.reference

    def __setitem__(self, key, value):
        super(AbstractResource, self).__setitem__(key, value)

    def __getitem__(self, key):
        try:
            return super(AbstractResource, self).__getitem__(key)
        except KeyError as e:
            raise AttributeError from e

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        try:
            super().__getattribute__(key)
            super().__setattr__(key, value)
        except AttributeError:
            self[key] = value

    def get_by_path(self, path, default=None):
        keys = parse_path(path)

        return get_by_path(self, keys, default)

    def get(self, key, default=None):
        return super(AbstractResource, self).get(key, default)

    def setdefault(self, key, default=None):
        return super(AbstractResource, self).setdefault(key, default)

    def serialize(self):
        def convert_fn(item):
            if isinstance(item, BaseResource):
                return item.to_reference().serialize(), True
            elif isinstance(item, BaseReference):
                return item.serialize(), True
            else:
                return item, False

        return convert_values({key: value for key, value in self.items()}, convert_fn)

    @property
    def id(self):  # pragma: no cover
        raise NotImplementedError()

    @property
    def resource_type(self):  # pragma: no cover
        raise NotImplementedError()

    @property
    def reference(self):  # pragma: no cover
        raise NotImplementedError()


class BaseResource(AbstractResource, ABC):
    resource_type = None

    def __init__(self, client, resource_type, **kwargs):
        def convert_fn(item):
            if isinstance(item, AbstractResource):
                return item, True

            if self.is_reference(item):
                return client.reference(**item), True

            return item, False

        self.resource_type = resource_type
        kwargs["resourceType"] = resource_type
        converted_kwargs = convert_values(kwargs, convert_fn)

        super(BaseResource, self).__init__(client, **converted_kwargs)

    def __setitem__(self, key, value):
        if key == "resourceType" and value != self.resource_type:
            raise KeyError(
                "Can not change `resourceType` after instantiating resource. "
                "You must re-instantiate resource using "
                "`Client.resource` method"
            )
        super(BaseResource, self).__setitem__(key, value)

    def __str__(self):  # pragma: no cover
        return "<{0} {1}>".format(self.__class__.__name__, self._get_path())

    def __repr__(self):  # pragma: no cover
        return self.__str__()

    @abstractmethod  # pragma: no cover
    def save(self, fields=None, search_params=None):
        pass

    @abstractmethod  # pragma: no cover
    def patch(self, **kwargs):
        pass

    @abstractmethod  # pragma: no cover
    def update(self):
        pass
    @abstractmethod  # pragma: no cover
    def delete(self):
        pass

    @abstractmethod  # pragma: no cover
    def refresh(self):
        pass

    def to_resource(self):
        """
        Returns Resource instance for this resource
        """
        return self

    def to_reference(self, **kwargs):
        """
        Returns Reference instance for this resource
        """
        if not self.reference:
            raise ResourceNotFound(
                "Can not get reference to unsaved resource without id"
            )

        return self.client.reference(reference=self.reference, **kwargs)

    @abstractmethod  # pragma: no cover
    def is_reference(self, value):
        pass

    @abstractmethod  # pragma: no cover
    def is_valid(self, raise_exception=False):
        pass

    @abstractmethod  # pragma: no cover
    def execute(self, operation, method=None, **kwargs):
        pass

    @property
    def id(self):
        return self.get("id", None)

    @property
    def reference(self):
        """
        Returns reference if local resource is saved
        """
        if self.id:
            return "{0}/{1}".format(self.resource_type, self.id)

    def _get_path(self):
        if self.id:
            return "{0}/{1}".format(self.resource_type, self.id)
        elif self.resource_type == "Bundle":
            return ""

        return self.resource_type


class BaseReference(AbstractResource):
    def __str__(self):  # pragma: no cover
        return "<{0} {1}>".format(self.__class__.__name__, self.reference)

    def __repr__(self):  # pragma: no cover
        return self.__str__()

    @abstractmethod  # pragma: no cover
    def to_resource(self):
        pass

    @abstractmethod  # pragma: no cover
    def execute(self, operation, method=None, **kwargs):
        pass

    def to_reference(self, **kwargs):
        """
        Returns Reference instance for this reference
        """
        return self.client.reference(reference=self.reference, **kwargs)

    def _dict_to_resource(self, data):
        return self.client.resource(data['resourceType'], **data)

    @property  # pragma: no cover
    @abstractmethod
    def reference(self):
        pass

    @property  # pragma: no cover
    @abstractmethod
    def id(self):
        """
        Returns id if reference specifies to the local resource
        """
        pass

    @property  # pragma: no cover
    @abstractmethod
    def resource_type(self):
        """
        Returns resource type if reference specifies to the local resource
        """
        pass

    @property  # pragma: no cover
    @abstractmethod
    def is_local(self):
        pass



from .lib import (
    SyncClient,
    AsyncClient,
    SyncSearchSet,
    AsyncSearchSet,
    SyncResource,
    AsyncResource,
    SyncReference,
    AsyncReference,
)

# TODO: Remove in 1.2.0
SyncAbstractClient = SyncClient
AsyncAbstractClient = AsyncClient
